######################################################################
#                                                                    #
#     A Simple Object System to Demonstrate Features of Tcl 8.5      #
#                                                                    #
######################################################################
#
# Copyright (c) 2004  Donal K. Fellows
#
# This file is placed under the same terms as the Tcl code.  It's
# "license.terms" applies to this file in full.
#

package require Tcl 8.5
package provide oo 0.1

######################################################################
# This is the core infrastructure for objects
#
namespace eval ::oo {
    variable state
    array set state {}
    variable defaultMethods {get set unset methods}
    variable counter 0
    variable definedMethods
    array set definedMethods {}
    namespace eval construct {}
}
proc ::oo::NewObject {} {
    variable counter
    variable state
    variable defaultMethods
    variable definedMethods

    while {[info command ::ob[incr counter]] ne ""} {}
    set ob ::ob$counter
    set state($counter) [dict create this $ob {state reference} $counter]
    set definedMethods($counter) {}
    set map {}
    foreach method $defaultMethods {
	dict set map $method [list ::oo::${method}Impl $counter]
    }
    namespace ensemble create -command $ob -map $map
    trace add command $ob delete [list ::oo::NukeObject $counter]
    trace add command $ob rename [list ::oo::RenameObject $counter]
    return $ob
}
proc ::oo::NukeObject {this oldName newName op} {
    variable state
    variable definedMethods
    foreach impl $definedMethods($this) {
	catch {rename $impl {}}
    }
    unset state($this) definedMethods($this)
}
proc ::oo::RenameObject {this oldName newName op} {
    variable state
    dict set state($this) this [uplevel 1 [list namespace which $newName]]
}

#Core Method Implementations
proc ::oo::getImpl {this field} {
    variable state
    dict get $state($this) $field
}
proc ::oo::setImpl {this field value} {
    variable state
    dict set state($this) $field $value
}
proc ::oo::unsetImpl {this field} {
    variable state
    dict unset state($this) $field
}
proc ::oo::methodsImpl {this args} {
    variable state
    variable definedMethods
    set ob [dict get $state($this) this]
    set methods [namespace ensemble configure $ob -map]
    if {[llength $args] == 0} {
	return [dict keys $methods]
    }
    if {[llength $args] == 1} {
	return [dict exists $methods [lindex $args 0]]
    }
    if {[llength $args] != 3} {
	return -code error \
		"wrong # args: must be \"$ob methods ?name? ?arguments body?\""
    }
    lassign $args name arguments body
    set body1 [list set {current method} [namespace tail $name]]
    set body2 ";dict with ::oo::state($this) "
    set shortname [namespace tail $name]
    append name ___[namespace tail $ob]
    uplevel \#0 [list proc $name [linsert $arguments 0 this__object] \
	    $body1$body2[list $body]]
    dict set methods $shortname [list $name $this]
    namespace ensemble configure $ob -map $methods
    lappend definedMethods($this) $name
}

######################################################################
# Bootstrap the OO framework
#
rename [::oo::NewObject] Class
proc ::oo::MakeNormalMethod {obj ns name arg body} {
    set fullname ${ns}::$name
    $obj method $fullname $arg $body
}
proc ::oo::InitProp {obj name {value {}}} {
    $obj set $name $value
}
proc ::oo::Inherit {obj superclass} {
    set classes [$obj get classes]
    if {!($superclass in $classes)} {
	$obj set classes [lappend classes $superclass]
    }
    uplevel 1 [$superclass get definition]
}
Class method ::oo::define {classname classDefinition} {
    if {$classname eq $this} {
	error "may not override the class \"$this\""
    }
    uplevel 1 [list rename [NewObject] $classname]
    $classname set definition $classDefinition
    $classname method ::oo::new {args} {
	set object [::oo::NewObject]
	$object set class $this
	$object set classes [list $this]
	set ns [uplevel 1 {namespace current}]
	interp alias {} ::oo::construct::constructor {} \
		::oo::construct::method constructor
	interp alias {} ::oo::construct::destructor {} \
		::oo::construct::method destructor {}
	interp alias {} ::oo::construct::method {} \
		::oo::MakeNormalMethod $object $ns
	interp alias {} ::oo::construct::property {} ::oo::InitProp $object
	interp alias {} ::oo::construct::inherit {} ::oo::Inherit $object
	namespace eval ::oo::construct $definition
	$object method ::oo::delete {} {
	    if {[$this method destructor]} {
		$this destructor
	    }
	    rename $this {}
	}
	if {[$object method constructor]} {
	    uplevel 1 [list $object constructor {expand}$args]
	}
	return $object
    }
    return $classname
}

Class define Object {
    ## This method is tricky because we want to compare core object
    ## references instead of command names to avoid aliasing problems
    method eq {other__object} {
	expr {$this__object == [$other__object get {state reference}]}
    }
    method isa {{{some class}}} {
	expr {${some class} in $classes}
    }
    method < {} {
	return ${state reference}
    }
    method > {{{the state reference}}} {
	dict get $::oo::state(${the state reference}) this
    }
    # A simple refcounting scheme
    property refcount 0
    method addRef {} {
	incr refcount
    }
    method delRef {} {
	if {[incr refcount -1] <= 0} {
	    $this delete
	}
    }
}

# Magically splice Class to be an Object
Class set class Class
Class set classes {Class Object}
Class set definition {error "may not inherit from Class"}
Class method new args {error "may not use new on $this"}
Class method delete {} {rename $this {}}
Class method addRef {} {}
Class method delRef {} {}
Class method eq {other__object} {
    expr {$this__object == [$other__object get {state reference}]}
}
Class method isa {{{some class}}} {
    expr {${some class} in $classes}
}
Class method < {} {
    return ${state reference}
}
Class method > {{{the state reference}}} {
    dict get $::oo::state(${the state reference}) this
}

Class define List {
    inherit Object
    property content {}
    method add {args} {
	foreach obj $args {
	    lappend content [$obj <]
	    $obj addRef
	}
    }
    method length {} {
	llength $content
    }
    method index {idx} {
	$this > [lindex $content $idx]
    }
    method find {obj} {
	lsearch -exact $content [$obj <]
    }
    method remove {idx} {
	[$this > [lindex $content $idx]] delRef
	set content [lreplace $content[set content {}] $idx $idx]
    }
    method iterate {var body} {
	set body "set [list $var] \[[list $this] > $[list $var]]$body"
	uplevel 1 [list foreach $var $content $body]
    }
    destructor {
	foreach ref $content {
	    [$this > $ref] delRef
	}
    }
}
Class define UniqueList {
    inherit List
    method add {args} {
	foreach obj $args {
	    if {[$this find $obj] == -1} {
		lappend content [$obj <]
		$obj addRef
	    }
	}
    }
}
